/* =============================================================================================================
	SyncthingIndicator 0.10
================================================================================================================

	GJS Syncthing gnome-shell panel indicator signalling the Syncthing deamon status.

	Credits to <jay.strict@posteo.de> for the reference implementation

	Copyright (c) 2018, 2nv2u <info@2nv2u.com>
	This work is distributed under GPLv3, see LICENSE for more information.
============================================================================================================= */

const Lang = imports.lang;
const Gio = imports.gi.Gio;
const Gtk = imports.gi.Gtk;
const St = imports.gi.St;

const Main = imports.ui.main;
const Animation = imports.ui.animation;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;
const StatusSystem = imports.ui.status.system;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();

const Logger = Me.imports.logger;
const console = new Logger.Service(Logger.Level.WARN,'SYNCTHING-UI');

const Syncthing = Me.imports.syncthing;

// Syncthing indicator panel icon
const SyncthingPanelIcon = new Lang.Class({
	Name: 'SyncthingPanelIcon',

	_init: function() {
		let scaleFactor = St.ThemeContext.get_for_stage(global.stage).scale_factor;
		this._workingIcon = new Animation.Animation(
			Gio.File.new_for_path(Me.path+'/icons/syncthing-working.svg'), 20, 20, 80
		);
		this._idleIcon = St.TextureCache.get_default().load_file_async(
			Gio.File.new_for_path(Me.path+'/icons/syncthing-idle.svg'), 20, 20, scaleFactor
		);
		this._pausedIcon = St.TextureCache.get_default().load_file_async(
			Gio.File.new_for_path(Me.path+'/icons/syncthing-paused.svg'), 20, 20, scaleFactor
		);
		this._disconnectedIcon = St.TextureCache.get_default().load_file_async(
			Gio.File.new_for_path(Me.path+'/icons/syncthing-disconnected.svg'), 20, 20, scaleFactor
		);
		this.actor = new St.Bin();
		this.actor.set_child(this._disconnectedIcon);
	},

	setState: function(state){
		this._workingIcon.stop();
		switch(state){
			case Syncthing.State.SYNCING:
			case Syncthing.State.SCANNING:
				this.actor.set_child(this._workingIcon.actor);
				this._workingIcon.play();
			break
			case Syncthing.State.PAUSED:
				this.actor.set_child(this._pausedIcon);
			break;
			case Syncthing.State.UNKNOWN:
			case Syncthing.State.DISCONNECTED:
				this.actor.set_child(this._disconnectedIcon);
			break;
			default:
				this.actor.set_child(this._idleIcon);
			break
		}
	}

});

// Syncthing indicator rescan menu item
const RescanMenuItem = new Lang.Class({
	Name: 'RescanMenuItem',
	Extends: PopupMenu.PopupBaseMenuItem,

	_init: function () {
		this.parent();
		this.setSensitive(false);
		this.actor.add_child(
			new St.Icon({
				gicon: new Gio.ThemedIcon({ name: 'emblem-synchronizing-symbolic' }),
				style_class: 'popup-menu-icon'
			})
		);

		this._label = new St.Label({ text: 'Rescan All Folders' });
		this.actor.add_child(this._label);
		this.actor.label_actor = this._label;
		this.actor.add_style_class_name('syncthing-button-spacing');
	},

	activate: function(event) {
		syncthingManager.rescan();
	}

});

// Syncthing indicator section menu
const SectionMenu = new Lang.Class({
	Name: 'SectionMenu',
	Extends: PopupMenu.PopupSubMenuMenuItem,

	_init: function(title,icon){
		this.parent(title,true);
		this.icon.icon_name = icon;
		this.icon.add_style_class_name('syncthing-submenu-icon');
		this.section = new PopupMenu.PopupMenuSection();
		this.section.actor.add_style_class_name('syncthing-menu-spacing');
		this.menu.addMenuItem(this.section);
	}

});

// Syncthing indicator fodler menu
const FolderMenu = new Lang.Class({
	Name: 'FolderMenu',
	Extends: SectionMenu,

	_init: function(){
		this.parent('Folders','system-file-manager-symbolic');

    this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());

		this.rescan = new RescanMenuItem();
		this.menu.addMenuItem(this.rescan);

		syncthingManager.connect(Syncthing.Signal.SERVICE_CHANGE, Lang.bind(this,function(manager,state){
			switch(state){
				case Syncthing.ServiceState.ACTIVE:
					this.rescan.setSensitive(true);
				break;
				case Syncthing.ServiceState.STOPPED:
					this.rescan.setSensitive(false);
				break;
			}
		}));
	}

});

// Syncthing indicator fodler menu item
const FolderMenuItem = new Lang.Class({
	Name: 'FolderMenuItem',
	Extends: PopupMenu.PopupBaseMenuItem,

	_init: function(folder) {
		this.parent();
		this._folder = folder;

		let icon;
		let file = Gio.File.new_for_path(this._folder.path);
		try {
			icon = file.query_info('standard::symbolic-icon', 0, null).get_symbolic_icon();
		} catch(e){
			if(e instanceof Gio.IOErrorEnum){
			  if(!file.is_native()){
				  icon = new Gio.ThemedIcon({ name: 'folder-remote-symbolic' });
			  } else {
				  icon = new Gio.ThemedIcon({ name: 'folder-symbolic' });
			  }
			} else {e
			   throw e;
			}
		}

		this.icon = new St.Icon({ gicon: icon, style_class: 'popup-menu-icon' });
		this.actor.add_child(this.icon);

		this.label = new St.Label({ text: folder.name });
		this.actor.add_child(this.label);
		this.actor.label_actor = this.label;

		this.path = file.get_uri();

		this._folder.connect(Syncthing.Signal.STATE_CHANGE, Lang.bind(this,function(folder,state){
			this.icon.style_class = 'popup-menu-icon '+state;
		}));

		this._folder.connect(Syncthing.Signal.DESTROY, Lang.bind(this,function(folder){
			this.destroy();
		}));

	},

	activate: function(event) {
		try {
			let launchContext = global.create_app_launch_context(event.get_time(), -1);
			Gio.AppInfo.launch_default_for_uri(this.path, launchContext);
		} catch(e) {
			Main.notifyError('Failed to launch URI: '+uri, e.message);
		}
		this.parent(event);
	}

});

// Syncthing indicator device menu
const DeviceMenu = new Lang.Class({
	Name: 'DeviceMenu',
	Extends: SectionMenu,

	_init: function(){
		this.parent('This device','computer-symbolic');

    this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());

    this._serviceSwitch = new PopupMenu.PopupSwitchMenuItem("Service", false, null);
    this._serviceSwitch.connect('activate', function(){
			return function(actor,event){
				if(actor.state) {
					syncthingManager.startService();
				} else {
					syncthingManager.stopService();
				}
			}
		}());
		this.menu.addMenuItem(this._serviceSwitch);

    this._autoSwitch = new PopupMenu.PopupSwitchMenuItem("Autostart", false, null)
    this._autoSwitch.connect('activate', function(){
			return function(actor,event){
				if(actor.state) {
					syncthingManager.enableService();
				} else {
					syncthingManager.disableService();
				}
			}
		}());
		this.menu.addMenuItem(this._autoSwitch);

    this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());

		this.config = new ConfigMenuItem();
		this.menu.addMenuItem(this.config);

		syncthingManager.connect(Syncthing.Signal.SERVICE_CHANGE, Lang.bind(this,function(manager,state){
			switch(state){
				case Syncthing.ServiceState.ACTIVE:
					this._serviceSwitch.setToggleState(true);
					this.config.setSensitive(true);
				break;
				case Syncthing.ServiceState.STOPPED:
					this._serviceSwitch.setToggleState(false);
					this.config.setSensitive(false);
				break;
				case Syncthing.ServiceState.ENABLED:
					this._autoSwitch.setToggleState(true);
				break;
				case Syncthing.ServiceState.DISABLED:
					this._autoSwitch.setToggleState(false);
				break;
			}
		}));

		syncthingManager.connect(Syncthing.Signal.ERROR, Lang.bind(this,function(manager,error){
			switch(error.type){
				case Syncthing.Error.DAEMON:
					this._serviceSwitch.setToggleState(false);
				break;
			}
		}));

	},

	setDevice: function(device){
		this.device = device;
		this.label.text = device.name;
		this.device.connect(Syncthing.Signal.STATE_CHANGE, Lang.bind(this,function(device,state){
			this.icon.style_class = 'popup-menu-icon syncthing-submenu-icon '+state;
		}));
	}

});

// Syncthing indicator device menu item
const DeviceMenuItem = new Lang.Class({
	Name: 'DeviceMenuItem',
	Extends: PopupMenu.PopupSwitchMenuItem,

	_init: function(device) {
		this.parent(device.name, false, null);

		this._device = device;

		let icon = new Gio.ThemedIcon({ name: 'network-server-symbolic' });
		this.icon = new St.Icon({ gicon: icon, style_class: 'popup-menu-icon' });
		this.actor.insert_child_at_index(this.icon,1);

		this.setSensitive(false);

		this._device.connect(Syncthing.Signal.STATE_CHANGE, Lang.bind(this,function(device){
			let state = device.getState()
			switch(state){
				case Syncthing.State.DISCONNECTED:
					this.setSensitive(false);
					this.setToggleState(false);
				break;
				case Syncthing.State.PAUSED:
					this.setSensitive(true);
					this.setToggleState(false);
				break;
				default:
					this.setSensitive(true);
					this.setToggleState(true);
				break
			}
			this.icon.style_class = 'popup-menu-icon '+state;
		}));

		this._device.connect(Syncthing.Signal.DESTROY, Lang.bind(this,function(){
			this.destroy();
		}));

		this.connect('activate', Lang.bind(this,function(actor){
			if(actor.state) {
				this._device.resume();
			} else {
				this._device.pause();
			}
		}));
	}

});

// Syncthing indicator config menu item
const ConfigMenuItem = new Lang.Class({
	Name: 'ConfigMenuItem',
	Extends: PopupMenu.PopupBaseMenuItem,

	_init: function(){
		this.parent();
		this.setSensitive(false);

		this.actor.add_child(
			new St.Icon({
				gicon: new Gio.ThemedIcon({ name: 'preferences-system-symbolic' }),
				style_class: 'popup-menu-icon'
			})
		);
		this.actor.add_child(
			new St.Label({ text: 'Open Web Interface' })
		);

		this.actor.add_style_class_name('syncthing-button-spacing');
	},

	activate: function(event) {
		let launchContext = global.create_app_launch_context(event.get_time(), -1);
		try {
			Gio.AppInfo.launch_default_for_uri(syncthingManager.getConfig().getURI(), launchContext);
		} catch(e) {
			Main.notifyError('Failed to launch URI: '+uri, e.message);
		}
	}

});

// Syncthing indicator controller
const SyncthingIndicator = new Lang.Class({
	Name: 'SyncthingIndicator',
	Extends: PanelMenu.Button,

	_init: function() {
		this.parent(0.0, "SyncthingIndicator");

		this.icon = new SyncthingPanelIcon();
		this.actor.add_actor(this.icon.actor);

		this.deviceMenu = new DeviceMenu();
		this.menu.addMenuItem(this.deviceMenu);

		this.folderMenu = new FolderMenu();
		this.menu.addMenuItem(this.folderMenu);

		this.defaultMenu = this.deviceMenu;

		this.menu.connect('open-state-changed', Lang.bind(this, function(menu,state){
			if(state) this.defaultMenu.menu.open();
		}));

		syncthingManager.connect(Syncthing.Signal.ERROR, Lang.bind(this,function(manager,error){
			switch(error.type){
				case Syncthing.Error.DAEMON:
					Main.notifyError('Syncthing indicator daemon error, '+error.message);
				break;
				case Syncthing.Error.SERVICE:
					Main.notifyError('Syncthing indicator service error, '+error.message);
				break;
				case Syncthing.Error.STREAM:
					Main.notifyError('Syncthing indicator decoding error, '+error.message);
				break;
				case Syncthing.Error.CONNECTION:
					Main.notifyError('Syncthing indicator connection error, '+error.message);
				break;
				case Syncthing.Error.CONFIG:
					Main.notifyError('Syncthing indicator config error, '+error.message);
				break;
			}
		}));

		syncthingManager.connect(Syncthing.Signal.SERVICE_CHANGE, Lang.bind(this,function(manager,state){
			switch(state){
				case Syncthing.ServiceState.ACTIVE:
					this.defaultMenu = this.folderMenu;
				break;
				case Syncthing.ServiceState.STOPPED:
					this.defaultMenu = this.deviceMenu;
				break;
			}
		}));

		syncthingManager.connect(Syncthing.Signal.FOLDER_ADD, Lang.bind(this,function(manager,folder){
			this.folderMenu.setSensitive(true);
			this.folderMenu.section.addMenuItem(
				new FolderMenuItem(folder)
			);
		}));

		syncthingManager.connect(Syncthing.Signal.DEVICE_ADD, Lang.bind(this,function(manager,device){
			this.deviceMenu.section.addMenuItem(
				new DeviceMenuItem(device)
			);
		}));

		syncthingManager.connect(Syncthing.Signal.HOST_ADD, Lang.bind(this,function(manager,device){
			this.deviceMenu.setDevice(device);
			device.connect(Syncthing.Signal.STATE_CHANGE, Lang.bind(this,function(device,state){
				this.icon.setState(state);
			}));
		}));
	}

});

function init(extension) {
	Gtk.IconTheme.get_default().append_search_path(extension.path + '/icons');
}

let syncthingManager;
let syncthingIndicator;

// Syncthing indicator main function / initiator
function enable() {
	syncthingManager = new Syncthing.Manager();
	syncthingIndicator = new SyncthingIndicator();
	Main.panel.addToStatusArea('syncthingIndicator', syncthingIndicator);
	syncthingManager.attach();
}

// Syncthing indicator main remove function
function disable() {
	syncthingIndicator.destroy();
	syncthingManager.destroy();
}
